template = """
    <!doctype html>
    <html class="no-js" lang="">
        <head>
            <meta charset="utf-8">
            <meta http-equiv="x-ua-compatible" content="ie=edge">
            <title>{{ service_name }}</title>
            <meta name="description" content="Landing">
            <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
            <link href="https://fonts.googleapis.com/css?family=Open+Sans:300,400,600,700" rel="stylesheet">
        </head>
        <style>
            html {
                font-family: sans-serif;
                -ms-text-size-adjust: 100%;
                -webkit-text-size-adjust: 100%;
                font-size: 10px;
                -webkit-tap-highlight-color: rgba(0, 0, 0, 0);
            }
            body {
                margin: 0;
                font-family: "Open Sans", sans-serif;
                font-size: 14px;
                line-height: 1.42857143;
                color: #333333;
                background-color: #fff;
            }
            a {
                background-color: transparent;
                color: #337ab7;
                text-decoration: none;
            }

            h1 {
                font-weight: 500;
                line-height: 1.1;
                color: inherit;
                text-align: center;
                padding-top: 2%;
                padding-bottom: 2%;
                text-transform: capitalize;
            }
            h3 {
                text-align: center;
            }
            iframe {
                width: 100%;
                height: 300px;
                border-style: solid;
                border-width: 1px;
                background-color: #E8E8E8;
            }
            .section {
                margin: 0 100px;
            }
        </style>
        <body>
            <h1>{{ service_name }}</h1>
            <h3>{{ version }}</h3>
            {%-if description -%}
                <div class="section">
                    <h2>Description</h2>
                    <p>{{ description }}</p>
                </div>
            {%- endif -%}
            <div class="section">
                <h2><a href="api/health">Health</a></h2>
                <iframe src="/api/health"></iframe>
            </div>
            {%- for swagger_version in swagger_versions -%}
                <div class="section">
                    <h2><a href="api/{{ swagger_version }}/swagger">Swagger ({{ swagger_version }})</a></h2>
                </div>
            {%- endfor -%}
            {%-if homepage -%}
                <div class="section">
                    <h2><a href={{ homepage }}>Home Page</a></h2>
                </div>
            {%- endif -%}
            <div class="section">
                <h2><a href="api/config">Config</a></h2>
                <iframe src="/api/config"></iframe>
            </div>
        </body>
    </html>
"""



"""
Generalized error handling.

"""
from logging import getLogger

from marshmallow import fields, Schema
from werkzeug.exceptions import default_exceptions

from microcosm_flask.conventions.encoding import dump_response_data


error_logger = getLogger("errors")


class SubErrorSchema(Schema):
    message = fields.String(required=True)


class ErrorContextSchema(Schema):
    errors = fields.List(fields.Nested(SubErrorSchema), required=True)


class ErrorSchema(Schema):
    message = fields.String(required=True, default="Unknown Error")
    code = fields.Integer(required=True, default=500)
    retryable = fields.Boolean(required=True, default=False)
    context = fields.Nested(ErrorContextSchema, required=False)


def as_retryable(error):
    """
    Given an exception, mark it as retryable when serializing
    into HTTP error response.

    """
    error.retryable = True

    return error


def extract_status_code(error):
    """
    Extract an error code from a message.

    """
    try:
        return int(error.code)
    except (AttributeError, TypeError, ValueError):
        try:
            return int(error.status_code)
        except (AttributeError, TypeError, ValueError):
            try:
                return int(error.errno)
            except (AttributeError, TypeError, ValueError):
                return 500


def extract_error_message(error):
    """
    Extract a useful message from an error.

    Prefer the description attribute, then the message attribute, then
    the errors string conversion. In each case, fall back to the error class's
    name in the event that the attribute value was set to a uselessly empty string.

    """
    try:
        return error.description or error.__class__.__name__
    except AttributeError:
        try:
            return str(error.message) or error.__class__.__name__
        except AttributeError:
            return str(error) or error.__class__.__name__


def extract_context(error):
    """
    Extract extract context from an error.

    Errors may (optionally) provide a context attribute which will be encoded
    in the response.

    """
    return getattr(error, "context", {"errors": []})


def extract_retryable(error):
    """
    Extract a retryable status from an error.

    It's not usually helpful to retry on an error, but it's useful to do so
    when the application knows it might.

    """
    return getattr(error, "retryable", False)


def extract_headers(error):
    """
    Extract HTTP headers to include in response.

    """
    return getattr(error, "headers",  {})


def extract_include_stack_trace(error):
    """
    Extract whether error should include a stack trace.

    """
    return getattr(error, "include_stack_trace", True)


def make_json_error(error):
    """
    Handle errors by logging and
    """
    message = extract_error_message(error)
    status_code = extract_status_code(error)
    context = extract_context(error)
    retryable = extract_retryable(error)
    headers = extract_headers(error)

    # Flask will not log user exception (fortunately), but will log an error
    # for exceptions that escape out of the application entirely (e.g. if the
    # error handler raises an error)
    error_logger.debug("Handling {} error: {}".format(
        status_code,
        message,
    ))

    # Serialize into JSON response
    response_data = {
        "code": status_code,
        "context": context,
        "message": message,
        "retryable": retryable,
    }
    # Don't pass in the error schema because it will suppress any extra fields
    return dump_response_data(None, response_data, status_code, headers)


def configure_error_handlers(graph):
    """
    Register error handlers.

    """
    # override all of the werkzeug HTTPExceptions
    for code in default_exceptions.keys():
        graph.flask.register_error_handler(code, make_json_error)

    # register catch all for user exceptions
    graph.flask.register_error_handler(Exception, make_json_error)



"""
Support running the development server.

"""
from argparse import ArgumentParser

from microcosm_flask.profiling import enable_profiling


def parse_args(graph):
    default_port = graph.config.flask.port
    default_enable_profiling = graph.config.flask.enable_profiling

    parser = ArgumentParser()
    parser.add_argument("--host", default="127.0.0.1")
    parser.add_argument("--port", type=int, default=default_port)
    parser.add_argument("--with-profiling", action="store_true", default=default_enable_profiling)
    parser.add_argument("--debug", action="store_true", default=False)
    return parser.parse_args()


def main(graph):
    args = parse_args(graph)

    if args.with_profiling:
        enable_profiling(graph)

    if args.debug:
        graph.metadata.debug = True

    try:
        graph.flask.run(host=args.host, port=args.port)
    except KeyboardInterrupt:
        pass



"""
Landing Page convention.

"""
from microcosm_flask.conventions.registry import iter_endpoints
from microcosm_flask.templates.landing import template

from distutils import dist
from io import StringIO
from jinja2 import Template
from pkg_resources import get_distribution, DistributionNotFound


def configure_landing(graph):

    def get_properties_and_version():
        """
        Parse the properties from the package information

        """
        try:
            distribution = get_distribution(graph.metadata.name)
            metadata_str = distribution.get_metadata(distribution.PKG_INFO)
            package_info = dist.DistributionMetadata()
            package_info.read_pkg_file(StringIO(metadata_str))
            return package_info
        except DistributionNotFound:
            return None

    def get_description(properties):
        """
        Calculate the description based on the package properties

        """
        if not properties:
            return None

        return '. '.join([
            field
            for field in [properties.description, properties.long_description]
            if field is not None
        ])

    def get_swagger_versions():
        """
        Finds all swagger conventions that are bound to the graph

        """
        versions = []

        def matches(operation, ns, rule):
            """
            Defines a condition to determine which endpoints are swagger type

            """
            if(ns.subject == graph.config.swagger_convention.name):
                return True
            return False

        for operation, ns, rule, func in iter_endpoints(graph, matches):
            versions.append(ns.version)

        return versions

    @graph.flask.route("/")
    def render_landing_page():
        """
        Render landing page

        """
        properties = get_properties_and_version()
        description = get_description(properties)
        swagger_versions = get_swagger_versions()

        return Template(template).render(
            service_name=graph.metadata.name,
            swagger_versions=swagger_versions,
            description=description,
            homepage=getattr(properties, 'url', None),
            version=getattr(properties, 'version', None),
        )



"""
A list field field that supports query string parameter parsing.

"""
from marshmallow.fields import List, ValidationError


class PrintableList(list):
    def __str__(self):
        return ",".join(str(item) for item in self)


class QueryStringList(List):
    def _deserialize(self, value, attr, obj):
        """
        _deserialize handles multiple formats of query string parameter lists
        including:

        /foo?bars=1,2
        /foo?bars[]=1&bars[]2

        and returns a list of values

        """
        if value is None:
            return None

        try:
            attribute_elements = [attr_element.split(",") for attr_element in obj.getlist(attr)]
            attribute_params = [param for attr_param in attribute_elements for param in attr_param]

            return PrintableList(super(QueryStringList, self)._deserialize(attribute_params, attr, obj))
        except ValueError:
            raise ValidationError("Invalid query string list argument")



"""
A URI valued (and normalized) field.

"""
from operator import itemgetter

from marshmallow.fields import Field, ValidationError
from rfc3986 import uri_reference


DEFAULT_PORTS = {
    "http": 80,
    "https": 443,
}


def normalize_uri(uri):
    """
    Normalize a URI (and return a string)

    """
    return normalize_uri_result(uri).unsplit()


def normalize_uri_result(uri):
    """
    Normalize a URI (And return a URIResult)

    """
    ref = uri_reference(uri).normalize()

    return ref._replace(
        authority=normalize_uri_authority(ref),
        query=normalize_uri_query(ref),
        path=normalize_uri_path(ref),
    )


def normalize_uri_query(ref):
    if not ref.query:
        return ref.query

    return "&".join([
        "=".join(term)
        for term
        in sorted(
            [
                term.split("=", 1)
                for term in ref.query.split("&")
            ],
            key=itemgetter(0),
        )
    ])


def normalize_uri_path(ref):
    if not ref.path:
        return ref.path
    return ref.path.rstrip("/")


def normalize_uri_authority(ref):
    if not ref.port:
        return ref.authority

    if DEFAULT_PORTS[ref.scheme] == int(ref.port):
        return ref.host

    return ref.authority


class URIField(Field):
    """
    Marshmallow field for a normalized URI.

    """
    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        return self.normalize(value)

    def _deserialize(self, value, attr, data):
        if value is None:
            return None
        return self.normalize(value)

    def normalize(self, value):
        result = normalize_uri_result(value)
        if not result.scheme:
            raise ValidationError("URI scheme is required for: {}".format(value))
        if not result.authority:
            raise ValidationError("URI authority is required for: {}".format(value))
        return result.unsplit()
